@@ -11,9 +11,13 @@ use Perinci::Access::InProcess;
 my $sah = Data::Sah->new();
 my $plc = $sah->get_compiler("perl");
 $plc->indent_character('');
-my $pa  = Perinci::Access::InProcess->new(load=>0, cache_size=>0);
+my $pa  = Perinci::Access::InProcess->new(
+    load               => 0,
+    cache_size         => 0,
+    extra_wrapper_args => {remove_internal_properties=>0},
+);
 
-our $VERSION = '0.01'; # VERSION
+our $VERSION = '0.02'; # VERSION
 
 use Moose;
 use namespace::autoclean;
@@ -72,22 +76,28 @@ sub munge_file {
     my $munged;
     my $in_pod;
     my ($pkg_name, $sub_name, $metas, $meta, $arg, $var);
-    my $sub_has_vargs; # VALIDATED_ARGS has been declared for the sub
-    my %vargs; # list of validated args for current sub
+    my $sub_has_vargs; # VALIDATE_ARGS has been declared for current sub
+    my %vargs; # list of validated args for current sub, val 2=skipped
     my %vsubs; # list of subs
     my $i = 0; # line number
 
     my $check_prev_sub = sub {
         return unless $sub_name;
         return unless $meta;
-        return unless $meta->{args};
         my %unvalidated;
         for (keys %{ $meta->{args} }) {
+            next unless $meta->{args}{$_}{schema};
             $unvalidated{$_}++ unless $vargs{$_};
         }
         if (keys %unvalidated) {
             $self->log("NOTICE: $fname: Some argument(s) not validated ".
                            "for sub $sub_name: ".join(", ", keys %unvalidated));
+        } elsif ((grep {$_==1} values %vargs) &&
+                     !defined($meta->{"_perinci.sub.wrapper.validate_args"})) {
+            $self->log(
+                "NOTICE: $fname: You might want to set ".
+                    "_perinci.sub.wrapper.validate_args => 0 in metadata ".
+                        "for sub $sub_name");
         }
     };
 
@@ -123,7 +133,7 @@ sub munge_file {
         push @code, 'my $arg_err; ' unless keys %vargs;
         push @code, __squish_code($cd->{result}), "; ";
         push @code, $gen_verr->('$arg_err', $arg);
-        $vargs{$arg}++;
+        $vargs{$arg} = 1;
         join "", @code;
     };
 
@@ -153,7 +163,7 @@ sub munge_file {
                     comment     => 0,
                 );
                 push @code, 'my $arg_err; ' unless keys %vargs;
-                $vargs{$arg}++;
+                $vargs{$arg} = 1;
                 push @code, __squish_code($cd->{result}), "; ";
                 push @code, $gen_verr->('$arg_err', $arg);
             }
@@ -195,8 +205,7 @@ sub munge_file {
             $log->tracef("Found sub declaration %s", $1);
             unless ($pkg_name) {
                 $self->log_fatal(
-                    "$fname:$i: sub without package definition");
-                next;
+                    "$fname:$i: module does not have package definition");
             }
             $check_prev_sub->();
             $sub_name      = $1;
@@ -205,66 +214,66 @@ sub munge_file {
             $meta          = $metas->{$sub_name};
             next;
         }
-        if (/^
-             (?<code>\s* my \s+ (?<sigil>[\$@%]) (?<var>\w+) \b .+)
-             (?<tag>\#\s*VALIDATE_ARG(?<s> S)? (?: \s+ (?<var2>\w+))? \s*$)/x) {
-            $log->tracef("Found line with tag %s", $_);
+        if (/^\s*?
+             (?<code>\s* my \s+ (?<sigil>[\$@%]) (?<var>\w+) \b .+)?
+             (?<tag>\#\s*(?<no>NO_)?VALIDATE_ARG(?<s> S)?
+                 (?: \s+ (?<var2>\w+))? \s*$)/x) {
             my %m = %+;
+            $log->tracef("Found line with tag %s, m=%s", $_, \%m);
+            next if !$m{no} && !$m{code};
             $arg = $m{var2} // $m{var};
+            if ($m{no}) {
+                if ($m{s}) {
+                    %vargs = map {$_=>2} keys %{$meta->{args} // {}};
+                } else {
+                    $vargs{$arg} = 2;
+                }
+                next;
+            }
             $var = $m{sigil} . $m{var};
             unless ($sub_name) {
-                $self->log(
-                    "$fname:$i: # VALIDATE_ARG(S?) outside sub");
-                next;
+                $self->log_fatal("$fname:$i: # VALIDATE_ARG$m{s} outside sub");
             }
             unless ($meta) {
                 $self->log_fatal(
-                    "$fname:$i: # VALIDATE_ARG(S?) ".
-                        "but no metadata for sub $sub_name");
-                next;
+                    "$fname:$i: sub $sub_name does not have metadata");
             }
             if (($meta->{v} // 1.0) != 1.1) {
                 $self->log_fatal(
-                    "$fname:$i: # VALIDATE_ARG(S?) but ".
-                        "metadata is not v1.1 (only v1.1 is supported)");
-                next;
+                    "$fname:$i: metadata for sub $sub_name is not v1.1 ".
+                        "(currently only v1.1 is supported)");
             }
             if (($meta->{args_as} // "hash") !~ /^hash(ref)?$/) {
                 $self->log_fatal(
-                    "$fname:$i: # VALIDATE_ARG(S?) for sub $sub_name: ".
-                        "Sorry, only args_as=hash/hashref currently supported");
-                next;
+                    "$fname:$i: metadata for sub $sub_name: ".
+                        "args_as=$meta->{args_as} (sorry, currently only ".
+                            "args_as=hash/hashref supported)");
             }
-            if (($meta->{v} // 1.0) != 1.1) {
+            unless ($meta->{args}) {
                 $self->log_fatal(
-                    "$fname:$i: # VALIDATE_ARG(S?) but ".
-                        "metadata does not have args definition");
-                next;
+                    "$fname:$i: # metadata for sub $sub_name: ".
+                        "no args property defined");
             }
             if ($m{s} && $sub_has_vargs) {
                 $self->log_fatal(
                     "$fname:$i: multiple # VALIDATE_ARGS for sub $sub_name");
-                next;
             }
             if (!$m{s}) {
                 unless ($meta->{args}{$arg} && $meta->{args}{$arg}{schema}) {
                     $self->log_fatal(
-                        "$fname:$i: # VALIDATE_ARG for ".
+                        "$fname:$i: metadata for sub $sub_name: ".
                             "no schema for argument $arg");
-                    next;
                 }
             }
-            if ($m{s} && $m{sigil} ne '%') {
+            if ($m{s} && $m{sigil} !~ /[\$%]/) {
                 $self->log_fatal(
                     "$fname:$i: invalid variable $var ".
-                        "for # VALIDATE_ARGS, must be hash");
-                next;
+                        "for # VALIDATE_ARGS, must be hash/hashref");
             }
             if (!$m{s} && $m{sigil} ne '$') {
                 $self->log_fatal(
                     "$fname:$i: invalid variable $var ".
                         "for # VALIDATE_ARG, must be scalar");
-                next;
             }
 
             $munged++;
@@ -300,7 +309,7 @@ Dist::Zilla::Plugin::Rinci::Validate - Insert argument validator code in output
 
 =head1 VERSION
 
-version 0.01
+version 0.02
 
 =head1 SYNOPSIS
 
@@ -374,6 +383,18 @@ attribute into your function metadata, to instruct Perinci::Sub::Wrapper to skip
 generating argument validation code when your function is wrapped, as argument
 validation is already done by the generated code.
 
+If there is an unvalidated argument, this plugin will emit a warning notice. To
+skip validating an argument (silence the warning), you can use:
+
+ sub foo {
+     my %args = @_;
+     my $arg1 = $args{arg1}; # NO_VALIDATE_ARG
+
+or:
+
+ sub foo {
+     # NO_VALIDATE_ARGS
+
 =head1 FAQ
 
 =head2 Rationale for this plugin?
